/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */



/* 
 * This file is part of the SINUS firewall loadable driver module
 *
 * $Id: sf_filter.c,v 1.40 1995/09/12 15:08:59 robby Rel $
 *
 */

#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS 1
#endif
#include <linux/modversions.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/icmp.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/ip.h>
#include <linux/malloc.h>
#include <linux/timer.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <linux/igmp.h>
#include <asm/checksum.h>
#include "sf_kernel.h"
#include "sf_rc.h"
#include "sf_global.h"
#include "sf_filter.h"
#include "sf_control.h"
#include "sf_tcp.h"
#include "sf_icmp.h"

/* loopback is the first device in the dev chain */
#define LOOPBACK_MASK (dev_base->pa_mask)
#define LOOPBACK_ADDR (dev_base->pa_addr)

/* firewall enabled */
int sf_fw_enabled = 0;

/* firewall initialized */
int sf_fw_initialized = 0;

/* temporary storage for packets to be sent to the daemon */
struct sf_proc sf_log_data;

/* wait queue for synchronization between filter function and daemon */
struct wait_queue *sf_log_entry = NULL;

#define PRINTIP(a) printk("%lu.%lu.%lu.%lu", ntohl(a)>>24&0xFF,\
                          ntohl(a)>>16&0xFF, ntohl(a)>>8&0xFF, ntohl(a)&0xFF)

#define MIN(a,b) ((a) < (b) ? (a) : (b))

/*
 * linear lists to store the rules and connections
 */

struct sf_fw *prio_rules = NULL;
struct sf_fw *rules      = NULL;
struct sf_tcp_connection *conns = NULL;
struct sf_tcp_connection *lastc = NULL;

/*
 * pointers to save positions for sf_****_first() and sf_****_next()
 * (used by "sfc show")
 */

struct sf_fw *rule_first_next = NULL;
unsigned long conn_id = 0;

/* 
 * array to store the ip addresses of the rules 
 */

struct sf_address sf_addr[SF_ADDRESS_CNT_MAX];
int sf_addr_free = 1;		/* first free entry in array */


/* returns 1 if addr matches one of the "internalnet" addresses */

static inline int sf_inside(__u32 addr)
{
  int i;

  for (i = 1; i <= sf_addr[0].addr; i++) {
    if ((sf_addr[i].addr & sf_addr[i].mask) == (addr & sf_addr[i].mask))
      return 1;
  }
  return 0;
}


/* returns 1 if address is local
 * dev_base is taken from <linux/netdevice.h> */

int sf_local(__u32 addr)
{
  struct device *pdev;
  
  for(pdev = dev_base; pdev; pdev = pdev->next) {
    if(pdev->pa_addr == addr)
      return 1;
  }
  return 0;
}

/* returns 1 if address matches */
int sf_addr_match(long idx, long cnt, __u32 addr, int port, int negative)
{
  int i;

  for (i = idx; i < (idx + cnt); i++) {		
    /* set addrmatch if address does not matter or if address matches */

    /* address is to be ignored if mask == 0 */
    int addrmatch =
    ((sf_addr[i].addr & sf_addr[i].mask) == (addr & sf_addr[i].mask));
    int portmatch = 1;

    /* reset portmatch if port matters but does not match */
    if (port >= 0)
      if ((port < sf_addr[i].port) || (sf_addr[i].prend < port))
        portmatch = 0;

    /* If the negative flag is not set, we have got a match if at
       least once addrmatch and portmatch is set, so return 1. 
       If the negative flag is set, the return value must be zero
       to be good. It may only be zero if the port matches (or
       does not matter) and none of the addresses match. As the 
       negative flag is set only if the outside keyword is used,
       all port fields contain the same value. So we can return
       one if portmatch is zero for the first array entry.       */
    if ((addrmatch && portmatch)
	|| (negative && !portmatch))
      return 1;

  }
  return 0;
}

/*
 * interface cache to determine if interface is connected to inside or outside.
 * Warning: there is collision detection, but no resolution; interface
 * base adresses should never collide.
 */

#define SF_IF_CACHE_SIZE	199	/* must be a prime number */
struct sf_if_cache_entry {
  __u32 ifaddr;			/* address of interface */
  int inside;			/* 1 if interface is inside */
};
struct sf_if_cache_entry sf_if_cache[SF_IF_CACHE_SIZE];

#define SF_IF_CACHE_KEY(a)  ( (a) % SF_IF_CACHE_SIZE )


#ifndef CONFIG_IP_ALWAYS_DEFRAG

/* 
 * hash queues for fragmented ip datagrams 
 */

struct sf_fragment {
  struct sf_fragment *prev, *next;	/* queue pointers */
  __u16 id;			/* identification field */
  __u32 src;			/* source address */
  __u32 dst;			/* destination address */
  struct timer_list timer;	/* expiration of datagram */
};

#define SF_FRAG_HASH_SIZE	499	/* must be a prime number */
struct sf_fragment *sf_frag[SF_FRAG_HASH_SIZE];

#define SF_FRAG_KEY(a)	( (a) % SF_FRAG_HASH_SIZE )

/* delete entry from hash queue */
static void sf_frag_expire(unsigned long arg)
{
  struct sf_fragment *frag;
  unsigned long flags;

  frag = (struct sf_fragment *) arg;
#ifdef SF_DEBUG_FRAG
  printk("sf_frag_expire: id = %u\n", ntohs(frag->id));
#endif
  del_timer(&frag->timer);
  save_flags(flags);
  cli();
  /* delete fragment from queue */
  if (frag->prev != NULL) {
    frag->prev->next = frag->next;
  }
  else
    sf_frag[SF_FRAG_KEY(frag->id)] = frag->next;
  if (frag->next != NULL)
    frag->next->prev = frag->prev;
  restore_flags(flags);
  kfree_s(frag, sizeof(struct sf_fragment));
}

/* insert hash queue entry */
static void sf_frag_insert(__u16 id, __u32 saddr, __u32 daddr)
{
  int key = SF_FRAG_KEY(id);
  struct sf_fragment *frag;
  unsigned long flags;

  /* allocate memory and initialize fragmentation hash queue entry */
  if ((frag =
       (struct sf_fragment *) kmalloc(sizeof(struct sf_fragment), GFP_ATOMIC)) == NULL) {
    printk("sf_frag_insert: out of memory!\n");
    return;
  }
  frag->id = id;
  frag->src = saddr;
  frag->dst = daddr;
#ifdef SF_DEBUG_FRAG
  printk("sf_frag_insert: id = %u\n", ntohs(frag->id));
#endif

  /* initialize timer */
  init_timer(&frag->timer);
  frag->timer.expires = jiffies + IP_FRAG_TIME;		/* same time as in reassembly */
  frag->timer.data = (unsigned long) frag;
  frag->timer.function = sf_frag_expire;
  add_timer(&frag->timer);

  /* insert as first element into hash queue */
  save_flags(flags);
  cli();
  if (sf_frag[key] != NULL)
    sf_frag[key]->prev = frag;
  frag->next = sf_frag[key];
  frag->prev = NULL;
  sf_frag[key] = frag;
  restore_flags(flags);
}

/* packet is a fragment if "more fragments" bit is set => add to hash queue */
static inline void check_for_fragment(struct iphdr *ip)
{
  if (ip->frag_off & htons(IP_MF)) 
    sf_frag_insert(ip->id, ip->saddr, ip->daddr);
}

#else

#define check_for_fragment(a)

#endif /* CONFIG_IP_ALWAYS_DEFRAG */


/*
 * hash queues for tcp circuits
 */

/* An established TCP connection is identified */
/* by the source and dest addresses and ports. */

#define SF_TCP_NO_SYN_YET	( 180 * HZ)	/* ftp data connection timeout */
#define SF_TCP_SYN_TO		( 120 * HZ)	/* connection establishment timeout */
#define SF_TCP_FIN_WAIT		(  15 * HZ)	/* wait after last FIN (let ACKs pass) */
#define SF_TCP_HALFCLOSE	( 480 * HZ)	/* half close timeout */

struct sf_tcp_connection *sf_tcp[SF_TCP_HASH_SIZE];

/* delete entry from hash queue */
typedef void (*sf_tcp_expire)(unsigned long);
static void sf_tcp_delete(struct sf_tcp_connection *sftcp)
{
  unsigned long flags;

  if (sftcp->timer_set) {
    del_timer(&sftcp->timer);
    sftcp->timer_set = 0;
  }

  save_flags(flags);
  cli();

  sftcp->id = 0;		/* invalidate */

  if (sftcp->prev != NULL) {
    sftcp->prev->next = sftcp->next;
  }
  else
    sf_tcp[SF_TCP_KEY(sftcp->client_addr + sftcp->server_addr
		      + (__u32) sftcp->client_port + (__u32) sftcp->server_port)] = sftcp->next;
  if (sftcp->next != NULL)
    sftcp->next->prev = sftcp->prev;

  if (sftcp->before != NULL)
    sftcp->before->after = sftcp->after;
  else
    conns = sftcp->after;
  if (sftcp->after != NULL)
    sftcp->after->before = sftcp->before;
  else
    lastc = sftcp->before;

#ifdef SF_DEBUG_TCP
  printk("sf: tcp: delete: %p ", sftcp); PRINTIP(sftcp->client_addr);
  printk(":%u -> ", ntohs(sftcp->client_port)); PRINTIP(sftcp->server_addr);
  printk(":%u\n", ntohs(sftcp->server_port));
#endif

  restore_flags(flags);
  kfree_s(sftcp, sizeof(struct sf_tcp_connection));
}

static void sf_tcp_reset(struct sf_tcp_connection *sftcp)
{
  struct iphdr *ip;
  struct tcphdr *tcp; 

  /* allocate minimum IP + minimum TCP header */
  ip  = kmalloc(sizeof(struct iphdr) + sizeof(struct tcphdr), GFP_ATOMIC); 
  tcp = (struct tcphdr *) ((char *) ip + sizeof(struct iphdr));
  memset(ip, 0, sizeof(struct iphdr) + sizeof(struct tcphdr));

  /* hand craft packet (checksums will be ignored) */
  ip->version  = 4;
  ip->ihl      = 5;
  ip->tos      = 0x10;
  ip->tot_len  = sizeof(struct iphdr) + sizeof(struct tcphdr);
  ip->id       = 0;
  ip->frag_off = 0;
  ip->ttl      = 255;
  ip->protocol = IPPROTO_TCP;
  ip->check    = 0;
  ip->saddr    = sftcp->client_addr; 
  ip->daddr    = sftcp->server_addr;

  tcp->source  = sftcp->client_port; 
  tcp->dest    = sftcp->server_port;
  tcp->seq     = sftcp->client_seq;

  send_tcp(1, sftcp->rif, ip, tcp);
  kfree(ip);

  sf_tcp_delete(sftcp);
}

static inline void set_tcp_timer(struct sf_tcp_connection *, unsigned long);
static inline void set_tcp_rst_timer(struct sf_tcp_connection *, unsigned long);

/* insert hash queue entry */
static struct sf_tcp_connection *sf_tcp_insert(__u32 saddr, __u32 daddr, 
                                               __u16 source, __u16 dest,
                                               struct sf_fw *rule,
					       struct device *rif)
{
  int key = SF_TCP_KEY(saddr + daddr + (__u32) source + (__u32) dest);
  struct sf_tcp_connection *sftcp;
  unsigned long flags;
  int sf_flush_all(void);

  /* allocate memory and initialize tcp hash queue entry */
  if ((sftcp = (struct sf_tcp_connection *) kmalloc(sizeof(struct sf_tcp_connection), GFP_ATOMIC)) == NULL) {
    printk("sf: tcp_insert: out of memory -- closing all connections!\n");

    /* Clean up the tcp hash queues. This will abort all TCP connections. */
    /* But if the hash queues occupy a lot of memory, it will help! */
    sf_flush_all();
    return NULL;
  }

  memset(sftcp, 0, sizeof(struct sf_tcp_connection));

  sftcp->client_addr  = saddr;
  sftcp->server_addr  = daddr;
  sftcp->client_port  = source;
  sftcp->server_port  = dest;
  sftcp->state        = SF_TCP_ACCEPT_SYN;
  sftcp->established  = jiffies;
  sftcp->rif          = rif;
  if (rule->timeout == SF_RULE_STATIC)
    sftcp->rule = rule;

#ifdef SF_DEBUG_TCP
  printk("sf: tcp: insert: (%i) %p ", key, sftcp); PRINTIP(sftcp->client_addr);
  printk(":%u -> ", ntohs(sftcp->client_port)); PRINTIP(sftcp->server_addr);
  printk(":%u\n", ntohs(sftcp->server_port));
#endif

  /* initialize timer */
  init_timer(&sftcp->timer);
  set_tcp_timer(sftcp, SF_TCP_NO_SYN_YET);
  sftcp->timer_set = 1;

  /* insert as first element into hash queue */
  save_flags(flags);
  cli();
  if (sf_tcp[key] != NULL)
    sf_tcp[key]->prev = sftcp;
  sftcp->next = sf_tcp[key];
  sftcp->prev = NULL;
  sf_tcp[key] = sftcp;

  /* make id, renumber ids if needed */
  if (lastc != NULL) {
    if (lastc->id == UINT_MAX) {
      /* renumber connection ids */
      struct sf_tcp_connection *ptr;
      __u32 id;

      ptr = conns;
      id = 1;
      while (ptr != NULL) {
	if (ptr->id == conn_id)
	  conn_id = id;
	ptr->id = id;
	id++;
	ptr = ptr->after;
      }
    }
    sftcp->id = lastc->id + 1;
  }
  else
    sftcp->id = 1;

  /* insert as last element into linked list of connections */
  if (conns == NULL)
    conns = sftcp;
  if (lastc != NULL)
    lastc->after = sftcp;
  sftcp->before = lastc;
  sftcp->after = NULL;
  lastc = sftcp;

  restore_flags(flags);
  return sftcp;
}

/*
 * RIP
 */

#define SF_RIP_PORT 520

struct riphdr {
  __u8 command;			/* RIP command */
  __u8 version;			/* RIP version */
  __u16 zero1;			/* must be zero */
};

struct ripdata {
  __u16 af;			/* address family */
  __u16 zero1;			/* must be zero */
  __u32 addr;			/* IP address */
  __u32 zero2;			/* must be zero */
  __u32 zero3;			/* must be zero */
  __u32 metric;			/* metric */
};

static inline int check_rip(struct sf_fw *rule, struct riphdr *rip, unsigned long riplength)
{
  unsigned long len;
  struct ripdata *rd = (struct ripdata *) (rip + 1);

#ifdef SF_DEBUG_RIP
  printk("RIP command %u version %u length %lu\n",
         (unsigned) rip->command,(unsigned) rip->version,riplength);
#endif

  for (len = riplength - sizeof(struct riphdr); len >= 20; len -= 20, rd++) {
#ifdef SF_DEBUG_RIP
    printk(" af %u addr ",ntohs(rd->af)); PRINTIP(rd->addr);
    printk(" metric %lu\n",ntohl(rd->metric));
#endif
    if ((ntohs(rd->af) == AF_INET)
	&& (ntohs(rd->metric) < 16)
	&& ((!(rule->fw_flags & SF_RIP_ADDR_NEG)
    && !sf_addr_match(rule->fw_rip_idx, rule->fw_rip_cnt, rd->addr, -1, 0))
	    || ((rule->fw_flags & SF_RIP_ADDR_NEG)
		&& sf_addr_match(rule->fw_rip_idx, rule->fw_rip_cnt, rd->addr, -1, 1))))
      return 0;
  }

  return 1;
}

static inline void set_tcp_timer(struct sf_tcp_connection *sftcp,
                            unsigned long when)
{
   if (sftcp->timer_set)
     del_timer(&sftcp->timer);

   sftcp->timer.expires  = jiffies + when;
   sftcp->timer.data     = (unsigned long) sftcp;
   sftcp->timer.function = (sf_tcp_expire) sf_tcp_delete;
   sftcp->timer_set      = 1;
   add_timer(&sftcp->timer);
}

static inline void set_tcp_rst_timer(struct sf_tcp_connection *sftcp,
                            unsigned long when)
{
   if (sftcp->timer_set)
     del_timer(&sftcp->timer);

   sftcp->timer.expires  = jiffies + when;
   sftcp->timer.data     = (unsigned long) sftcp;
   sftcp->timer.function = (sf_tcp_expire) sf_tcp_reset;
   sftcp->timer_set      = 1;
   add_timer(&sftcp->timer);
}


/* update usage counter, send packet to daemon, append to message buffer */
static int inform_log(struct sf_fw *rule, struct iphdr *ip, 
                      struct device *rif, int ip_tot_len)
{
  rule->usage++;

  /* no notification? */
  if (!(rule->fw_flags & SF_FW_LOG))
    return 0;

  if (sf_log_data.num_entries < NUM_PROC_ENTRIES) {
    struct sf_proc_entry *current_p = &(sf_log_data.data[sf_log_data.num_entries]);

    current_p->action     = rule->fw_rc;
    current_p->rule.num   = rule->rule_id.num;
    strncpy(current_p->devname, rif->name, 7);
    current_p->devname[7] = 0;
    memcpy(current_p->ip_packet, ip, MIN(PROC_SIZE_IP, ip_tot_len));
    sf_log_data.num_entries++;
  }
  else {
    sf_log_data.lost_entries++;

    /* wake up daemon */
    wake_up_interruptible(&sf_log_entry);

    /* drop packet silently if buffer is full */
    return -1;
  }

  /* wake up daemon */
  wake_up_interruptible(&sf_log_entry);

  return 0;
}


/*
 * Returns FW_BLOCK if packet should be dropped, FW_ACCEPT if it should be 
 * accepted, and FW_REJECT if an ICMP unreachable packet should be sent.
 *
 * opt is one of SF_STATE_FORWARD, TRANSMIT or RECEIVE.
 *
 * This call is done from task[0], so it must not sleep.  
 */

int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
{
  struct sf_fw implicit_rule;
  struct sf_fw *rule = &implicit_rule;
  int ip_hdr_len = ip->ihl * 4;
  int ip_tot_len = ntohs(ip->tot_len);
  char *payload  = (char *) ip + ip_hdr_len;
  char *pktend   = (char *) ip + ip_tot_len;
#define tcp  ((struct tcphdr  *) payload)
#define udp  ((struct udphdr  *) payload)
#define icmp ((struct icmphdr *) payload)
#define igmp ((struct igmphdr *) payload)
  int returncode  = FW_BLOCK;

  /* --- TCP variables --- */
  struct sf_tcp_connection *sftcp = NULL;
  char *tcpdata   = NULL;
  int sequence_ok = 0;
  int reverse     = 0;		/* indicates direction of packet */
    				/* reverse == 0: client->server  */
    				/* reverse == 1: client<-server  */
  int key; 
  __u32 end_seq = 0;
  __u32 ack_seq = 0;
  __u16 window  = 0;
  /* --- end TCP variables --- */

  /* don't let packet pass if firewall functions are disabled */
  if (sf_fw_enabled <= 0)
    return FW_BLOCK;

  /* check if packet is long enough to hold IP header and an IPv4 packet */
  if (ip_tot_len < ip_hdr_len || ip->ihl < 5 || ip->version != 4)
    return FW_BLOCK;

  rule->fw_rc    = FW_BLOCK;
  rule->fw_flags = SF_FW_LOG;

  /* --- check for loopback spoofing --- */
  /* if the device is loopback, source and dest addr. must be loopback addr. */
  if (!(rif->flags & IFF_LOOPBACK)
      && (((ip->saddr & LOOPBACK_MASK) ==
	   (LOOPBACK_ADDR & LOOPBACK_MASK))
	  || ((ip->daddr & LOOPBACK_MASK) ==
	      (LOOPBACK_ADDR & LOOPBACK_MASK)))) {	/* packet not allowed */
    rule->rule_id.num = (opt == SF_STATE_RECEIVE) 
                        ? RULE_SPOOF_RECV : RULE_SPOOF_XMIT;
    goto informlog;		/* skip all further examination but inform daemon */
  }
  /* --- end loopback check --- */


  /* check local packets only once -- at transmit time */
  if (opt == SF_STATE_RECEIVE && ip->saddr == ip->daddr)
    return FW_ACCEPT;

  /* don't check forwarded packets at transmit time */
  if (opt == SF_STATE_TRANSMIT && !sf_local(ip->saddr))
    return FW_ACCEPT;

  /* --- check interface --- */
  {
    int ifinside, ifkey;

    /* determine if interface is connected to inside or to outside */
    if (sf_if_cache[ifkey = SF_IF_CACHE_KEY(rif->pa_addr)].ifaddr == rif->pa_addr) {
      ifinside = sf_if_cache[ifkey].inside;
    }
    else {			/* interface not in cache */
#ifdef SF_DEBUG
      printk("sf: caching interface "); PRINTIP(rif->pa_addr);
      printk(", cache entry %i\n", ifkey);
#endif
      ifinside = sf_inside(rif->pa_addr);
      sf_if_cache[ifkey].ifaddr = rif->pa_addr;
      sf_if_cache[ifkey].inside = ifinside;
    }

    /* detect spoofing on non-loopback devices: source/dest interface
     * is internal, ip number is external or vice versa */
    if (!(rif->flags & IFF_LOOPBACK)
	&& (((opt == SF_STATE_RECEIVE)
	     && !(IN_CLASSD(ntohl(ip->saddr)))
	     && (ifinside != sf_inside(ip->saddr)))
	    || (((opt == SF_STATE_TRANSMIT)
		 || (opt == SF_STATE_FORWARD))
		&& !(IN_CLASSD(ntohl(ip->daddr)))
		&& (ifinside != sf_inside(ip->daddr))))) {	/* packet not allowed */

      rule->rule_id.num = (opt == SF_STATE_RECEIVE) 
                          ? RULE_SPOOF_RECV : RULE_SPOOF_XMIT;
      goto informlog;		/* skip all further examination but inform daemon */
    }
  }
  /* --- end check interface --- */


  /* If SF_STATE_FORWARD, the packet has already been examined at receive 
     time. So skip the examination now. Do NOT move this test before the
     interface check, because at SF_STATE_RECEIVE we examine a different
     interface. */
  if (opt == SF_STATE_FORWARD)
    return FW_ACCEPT;

  rule->rule_id.num = RULE_CONNECTION;

  /* --- check fragments --- */
  if (ip->frag_off & htons(IP_OFFSET)) {	/* fragment offset != 0 */
#ifndef CONFIG_IP_ALWAYS_DEFRAG
    /* this is an ip fragment but not the first one of the packet */
    struct sf_fragment *frag = sf_frag[SF_FRAG_KEY(ip->id)];

#ifdef SF_DEBUG_FRAG
    printk("sf_check_packet: frag_off %u  tot_len %u\n",
	   (ntohs(ip->frag_off) & IP_OFFSET) << 3, ip_tot_len);
#endif

    if (((ntohs(ip->frag_off) & IP_OFFSET) << 3) + ip_tot_len > 65535) {
#ifdef SF_DEBUG_FRAG
      printk("sf_check_packet: oversized packet detected!\n");
#endif
      rule->rule_id.num = RULE_OVERSIZED;
      goto informlog;		/* skip all further examination but inform daemon */
    }

    while (frag != NULL) {
      if ((ip->id == frag->id) && (ip->saddr == frag->src) && (ip->daddr == frag->dst)) {	/* fragment belongs to an allowed packet */
	/* Always let the timer delete the queue entry, because
	   fragments need not arrive in order, thus there may still
	   be fragments missing when the last fragment arrives. */

	del_timer(&frag->timer);
	frag->timer.expires  = jiffies + IP_FRAG_TIME;	/* same time as in reassembly */
	frag->timer.data     = (unsigned long) frag;
	frag->timer.function = sf_frag_expire;
	add_timer(&frag->timer);
	return FW_ACCEPT;
      }
      frag = frag->next;
    }
#endif /* CONFIG_IP_ALWAYS_DEFRAG */
    /* fragment must not pass */
    return FW_BLOCK;
  }
  /* --- end check fragments --- */


  /* Now we know that the packet is not a subsequent fragment.
     So we can check, if the packet is long enough to hold an
     entire UDP, ICMP, IGMP (or TCP, see below) header */

  if(ip->protocol == IPPROTO_UDP)
    if (ip_tot_len < ip_hdr_len + sizeof(struct udphdr))
       return FW_BLOCK;

  if(ip->protocol == IPPROTO_ICMP)
    if (ip_tot_len < ip_hdr_len + sizeof(struct icmphdr))
       return FW_BLOCK;

  if(ip->protocol == IPPROTO_IGMP)
    if (ip_tot_len < ip_hdr_len + sizeof(struct igmphdr))
       return FW_BLOCK;

  /* check for land attack with UDP packets
     -- same in and out socket, see CERT adv. 79.28 */
  if(ip->protocol == IPPROTO_UDP &&
     ip->saddr == ip->daddr && udp->source == udp->dest)
    return FW_BLOCK;
  /* add Logfile entry sometimes -- hw */

  /* --- check tcp --- */
  if(ip->protocol == IPPROTO_TCP) {
    if(ip_tot_len < ip_hdr_len + sizeof(struct tcphdr))
       return FW_BLOCK;

    /* check for land attack with TCP packets
       -- same in and out socket, see CERT adv. 79.28 */
    if(ip->saddr == ip->daddr && tcp->source == tcp->dest)
      return FW_BLOCK;
    /* add Logfile entry sometimes -- hw */

    key = SF_TCP_KEY(ip->saddr + ip->daddr + (__u32) tcp->source + 
                     (__u32) tcp->dest);

#ifdef SF_DEBUG_TCP_VERBOSE
    printk("sf: tcp: "); PRINTIP(ip->saddr); 
    printk(":%u -> ", ntohs(tcp->source)); PRINTIP(ip->daddr);
    printk(":%u\n", ntohs(tcp->dest)); 
#endif

    /* convert these to host order for faster processing */
    tcpdata = payload + tcp->doff*4;
    end_seq = ntohl(tcp->seq) + pktend - tcpdata;
    ack_seq = ntohl(tcp->ack_seq);
    window  = ntohs(tcp->window);

    sftcp = sf_tcp[key];

    while (sftcp != NULL) {
      if ((sftcp->client_addr == ip->saddr) &&
	  (sftcp->client_port == tcp->source) &&
	  (sftcp->server_addr == ip->daddr) &&
	  (sftcp->server_port == tcp->dest)) {
	/* reverse = 0; --- packet from client to server [default] */
	goto have_connection;
      }
      if ((sftcp->client_addr == ip->daddr) &&
	  (sftcp->client_port == tcp->dest) &&
	  (sftcp->server_addr == ip->saddr) &&
	  (sftcp->server_port == tcp->source)) {
	reverse++;		/* packet from server to client */
	goto have_connection;
      }
      sftcp = sftcp->next;
    }

    /* assert(sftcp == NULL) */
 
    if (tcp->syn && !tcp->ack) {
      /* special case for passive mode ftp */

      sftcp = sf_tcp[SF_TCP_KEY(ip->saddr + ip->daddr + (__u32) tcp->dest)];

      while (sftcp != NULL) {
        if ((sftcp->client_addr == ip->saddr) &&
	    sftcp->allpsource &&
	    (sftcp->server_addr == ip->daddr) &&
	    (sftcp->server_port == tcp->dest)) {
	  struct sf_fw *rule = sftcp->rule;
          int log_open = sftcp->log_open;

	  /* reverse = 0; --- packet from client to server [default] */

          sf_tcp_delete(sftcp);
#ifdef SF_DEBUG_TCP
	  printk("sf: tcp: using wildcard connection ex %p\n", sftcp);
#endif
          sftcp = sf_tcp_insert(ip->saddr, ip->daddr, tcp->source, tcp->dest, rule, rif);
          if (sftcp != NULL) {
            sftcp->log_open = log_open;
	    goto have_connection;
          }
          else
            break;
        }
        sftcp = sftcp->next;
      }

      goto checkrules;	/* new connection */
    }


#ifdef SF_DEBUG_TCP
    printk("sf: tcp: connection not in hash queue\n");
#endif
    return FW_BLOCK;	/* block packet */


    /* connection found in hash queue: assert(sftcp != NULL) */
have_connection:
    if (sftcp->rule != NULL) {
      sftcp->rule->bytes += ip_tot_len;
      if (sftcp->rule->bytes >= GIGA) {
        sftcp->rule->gbytes++;
        sftcp->rule->bytes -= GIGA;
      }
    }

#ifdef SF_DEBUG_TCP
    if (tcp->syn || tcp->fin || tcp->rst) {
      printk("sf: tcp: %p %s", sftcp, reverse ? "s->c" : "c->s");
      if (tcp->syn) printk(" SYN");
      if (tcp->fin) printk(" FIN");
      if (tcp->ack) printk(" ACK");
      if (tcp->rst) printk(" RST");
      printk("\n");
      printk("sf: tcp: end_seq=%08x, ack_seq=%08x\n", end_seq, ack_seq);
    }
#endif

    sftcp->lastused = jiffies;


    /* --- test checksum --- */
#ifndef CONFIG_IP_ALWAYS_DEFRAG
    if ((ip->frag_off & htons(IP_MF)) == 0)
#endif
      if (tcp_check(tcp, pktend - payload, ip->saddr, ip->daddr,
          csum_partial(payload, pktend - payload, ~tcp->check)) != tcp->check) {
#ifdef SF_DEBUG_CSUM
        printk("sf: tcp: wrong checksum (hdr = %04hx)\n", tcp->check);
#endif

        return FW_BLOCK;
      }
    /* --- end test checksum --- */


    /* --- check sequence numbers --- */
    if (sftcp->check_sequence) {
      if (!tcp->syn) {
        if (!reverse) {
          /* OK = client_seq <= end_seq <= client_seq+server_win */
          if (between(end_seq, sftcp->client_seq-1, 
              sftcp->client_seq + (sftcp->server_win << sftcp->server_scale))) {
            sequence_ok++;
            sftcp->client_win = window;
          }
#ifdef SF_DEBUG_SEQ
          else
            printk("client_seq %u end_seq %u server_win %u client_win %u\n", 
                   sftcp->client_seq, end_seq, 
                   sftcp->server_win << sftcp->server_scale, 
                   sftcp->client_win << sftcp->client_scale);
#endif
        }
        else {
          /* OK = server_seq <= end_seq <= server_seq+client_win */
          if (between(end_seq, sftcp->server_seq-1,
              sftcp->server_seq + (sftcp->client_win << sftcp->client_scale))) {
            sequence_ok++;
            sftcp->server_win = window;
          }
#ifdef SF_DEBUG_SEQ
          else
            printk("server_seq %u end_seq %u client_win %u server_win %u\n", 
                   sftcp->server_seq, end_seq, 
                   sftcp->client_win << sftcp->client_scale, 
                   sftcp->server_win << sftcp->server_scale);
#endif
        }

        if (sftcp->state >= SF_TCP_ESTABLISHED3 && sequence_ok && tcp->ack) {
          /* update minimum sequence number */

          if (!reverse) {
            if (after(ack_seq, sftcp->server_seq-1))
              sftcp->server_seq = ack_seq;
#ifdef SF_DEBUG_SEQ
            else
              printk("server_seq %u ack_seq %u client_win %u server_win %u\n", 
                     sftcp->server_seq, ack_seq, 
                     sftcp->client_win << sftcp->client_scale, 
                     sftcp->server_win << sftcp->server_scale);
#endif
          }
          else {
            if (after(ack_seq, sftcp->client_seq-1))
              sftcp->client_seq = ack_seq;
#ifdef SF_DEBUG_SEQ
            else
              printk("client_seq %u ack_seq %u server_win %u client_win %u\n", 
                     sftcp->client_seq, ack_seq, 
                     sftcp->server_win << sftcp->server_scale, 
                     sftcp->client_win << sftcp->client_scale);
#endif
          }
        }
      } /* end !syn */
    }
    else /* ignore sequence numbers */
      sequence_ok++;
    /* --- end check sequence numbers --- */


    /* --- simple packets: no fin, no rst --- */
    if (!tcp->fin && !tcp->rst) {

      /* --- default case goes first --- */
      if (sftcp->state == SF_TCP_ESTABLISHED3) {
        /* This branch also catches SYN packets on an established connection. */

#ifdef SF_DEBUG_SEQ
        /* Wrong sequence numbers most likely result from resent packets */
        if (!sequence_ok)
          printk("sf: tcp: wrong sequence number [default]\n");
#endif

        check_for_fragment(ip);		/* check if packet is fragmented */
        return FW_ACCEPT;		/* packet is allowed */
      } 
      /* --- end default case --- */

      /* --- filter FTP --- */
      if (sftcp->state == SF_TCP_ESTABLISHEDFTP && sequence_ok) {

        if (!reverse) {
	  /* FTP packet from client to server */

#ifdef SF_DEBUG_FTP
	  char cmd[80];
	  char *p;
	  int cmdlen;

	  p = (char *) memscan(tcpdata, '\r', pktend - tcpdata);
	  cmdlen = (p < pktend) ? p - tcpdata : 0;
	  if (cmdlen > 0) {
	    strncpy(cmd, tcpdata, cmdlen);
	    cmd[cmdlen] = 0;
	    printk("sf ftp: c->s: %s\n", cmd);
	  }
#endif

	  /* Format of port command is "PORT n1,n2,n3,n4,n5,n6"
	     where n1.n2.n3.n4 is the IP address 
	     and n5*256+n6 is the port number.
	     Convert the string into numbers */

	  if ((sftcp->ftp_allow_active) &&
              (pktend - tcpdata > 15) && 
              (strncmp(tcpdata, "PORT ", 5) == 0)) {
	    int i;
	    __u16 ftpport = 0;
	    struct sf_tcp_connection *sfftp = NULL;
            char *ftpdata = tcpdata + 5;

	    /* Skip the IP address and use the packet source address
	       instead. So it is not possible to redirect the data 
	       connection to another host. */

	    for (i = 0; i < 4; i++) {
	      while (ftpdata < pktend && *ftpdata != ',')
	        ftpdata += 1;
	      ftpdata += 1;
	    }

	    /* convert port number */
	    for (i = 0; i < 2; i++) {
	      int j = 0;

	      ftpport <<= 8;

	      while (ftpdata < pktend && *ftpdata >= '0' && *ftpdata <= '9') {
	        j = j * 10 + *ftpdata - '0';
	        ftpdata += 1;
	      }

	      ftpdata += 1;
	      ftpport += (__u16) j;
	    }

	    if (((ftpport >= 1024) && (ftpport < 6000)) || (ftpport > 6100)) {
#ifdef SF_DEBUG_FTP
	      printk("sf ftp: accepting active port "); PRINTIP(ip->saddr);
	      printk(":%u\n", ftpport);
#endif
	      sfftp = sf_tcp_insert(ip->daddr, ip->saddr, htons(FTP_DATA_PORT), 
                                  htons(ftpport), sftcp->rule, rif);
              if (sfftp != NULL)
                sfftp->log_open = sftcp->log_dxfer;
            } /* end port >= 1024 and not X11 */
	    else {
#ifdef SF_DEBUG_FTP
	      printk("sf ftp: rejecting active port "); PRINTIP(ip->saddr);
	      printk(":%u\n", ftpport);
#endif
	
              /* Store ftpport so that a meaningful log message can be 
                 generated ("FTP: block TCP from dest:ftpdata to source:port") */
              rule->fw_rc       = htons(ftpport);
              rule->rule_id.num = RULE_REJECT_FTP;
              returncode        = FW_ACCEPT;
              goto informlog;
            }
	  } /* end PORT command detected */ 
        }
        else { /* reverse */
	  /* FTP packet from server to client */

#ifdef SF_DEBUG_FTP
	  char cmd[80];
	  char *p;
	  int cmdlen;

	  p = (char *) memscan(tcpdata, '\r', pktend - tcpdata);
	  cmdlen = (p < pktend) ? p - tcpdata : 0;
	  if (cmdlen > 0) {
	    strncpy(cmd, tcpdata, cmdlen);
	    cmd[cmdlen] = 0;
	    printk("sf ftp: s->c: %s\n", cmd);
	  }
#endif

	  /* Format of passive reply is "227 blah (n1,n2,n3,n4,n5,n6)" 
	     where the parantheses are optional, n1.n2.n3.n4 is the 
	     IP address and n5*256+n6 is the port number. */

	  if ((sftcp->ftp_allow_passive) &&
              (pktend - tcpdata > 14) && 
              (strncmp(tcpdata, "227 ", 4) == 0)) {
	    int i;
	    __u16 ftpport = 0;
	    struct sf_tcp_connection *sfftp = NULL;
            char *ftpdata = tcpdata + 5;

	    /* Skip the IP address and use the packet source address
	       instead. So it is not possible to redirect the data 
	       connection to another host. */

	    for (i = 0; i < 4; i++) {
	      while (ftpdata < pktend && *ftpdata != ',')
	        ftpdata++;
	      ftpdata++;
	    }

	    /* convert port number */
	    for (i = 0; i < 2; i++) {
	      int j = 0;

	      ftpport <<= 8;

	      while (ftpdata < pktend && *ftpdata >= '0' && *ftpdata <= '9') {
	        j = j * 10 + *ftpdata - '0';
	        ftpdata++;
	      }

	      ftpdata++;
	      ftpport += (__u16) j;
	    }

#ifdef SF_DEBUG_FTP
	    printk("sf ftp: accepting passive port "); PRINTIP(ip->daddr);
	    printk(":%u\n", ftpport);
#endif
	    sfftp = sf_tcp_insert(ip->daddr, ip->saddr, 0, htons(ftpport), sftcp->rule, rif);
	    if (sfftp != NULL) {
              sfftp->allpsource = 1;
              sfftp->log_open = sftcp->log_dxfer;
            }
	  } /* end passive 227 reply detected */
        }

        check_for_fragment(ip);	/* check if packet is fragmented */
        return FW_ACCEPT;	/* packet is allowed */
      }
      /* --- end FTP filter --- */

      /* --- this is similar to the default case, but more options --- */
      if (sftcp->state >= SF_TCP_ESTABLISHED3) {
#ifdef SF_DEBUG_SEQ
        if (!sequence_ok)
          printk("sf: tcp: wrong sequence number [extra]\n");
#endif

        if (sftcp->state > SF_TCP_ESTABLISHEDFTP &&
            sftcp->state < SF_TCP_TERMINATED) {
          /* support half close: reset timeout */

          set_tcp_timer(sftcp, SF_TCP_HALFCLOSE);

#ifdef SF_DEBUG_TCP
	  printk("sf: tcp: data in half close, resetting timeout\n");
#endif
        }

        check_for_fragment(ip);	/* check if packet is fragmented */
        return FW_ACCEPT;	/* packet is allowed */
      }
      /* --- end more options in state >= ESTABLISHED3 --- */

      /* --- ACK to complete 3-way handshake --- */
      if (sftcp->state == SF_TCP_SYN_ACK && !reverse) {
        /* We are waiting for ACK to complete 3-way handshake */

        if (sftcp->check_sequence) {
#ifdef SF_DEBUG_SEQ
          if (!sequence_ok)
            printk("sf: tcp: wrong sequence number [3-way-ack]\n");
#endif

          if (ack_seq-1 == sftcp->server_seq)
            sftcp->server_seq++;
          else {
#ifdef SF_DEBUG_SEQ
            if (!sequence_ok)
              printk("Are we guessing sequence numbers!?\n");
            printk("server_seq %u ack_seq %u server_win %u client_win %u\n", 
                    sftcp->server_seq, ack_seq, 
                    sftcp->server_win << sftcp->server_scale, 
                    sftcp->client_win << sftcp->client_scale);
#endif
            return FW_BLOCK;
          }
        }

        sftcp->state = SF_TCP_ESTABLISHED3;
        if (sftcp->server_port == htons(FTP_CONTROL_PORT))
          sftcp->state = SF_TCP_ESTABLISHEDFTP;

        del_timer(&sftcp->timer);
        sftcp->timer_set = 0;

        check_for_fragment(ip);
        return FW_ACCEPT;
      }
      /* --- end ACK to complete 3-way handshake --- */

      /* --- SYN: initiate new connection or step 2 in handshake --- */
      if (tcp->syn) {
        int opt_scale = -1; /* TCP window scaling option */
        char *tcpopt;

        if (sftcp->state >= SF_TCP_TERMINATED) {	
          /* Somebody is trying to establish a connection that is still in the 
             2 MSL wait state. This should not happen if the TCP protocol is 
             implemented correctly. But if it happens nevertheless, we must 
             not allow the connection without checking the rules. 

             See also: RFC 1185 */

          if (sftcp->check_sequence && 
              (!after(end_seq, reverse ? sftcp->server_seq : sftcp->client_seq)))
            return FW_BLOCK;

#ifdef SF_DEBUG_TCP
	  printk("sf: tcp: SYN && SF_TCP_TERMINATED\n");
#endif
	  sf_tcp_delete(sftcp);
          sftcp = NULL;
          goto checkrules;
        }

        if (sftcp->check_sequence) {
          /* parse TCP options */
          tcpopt = payload + sizeof(struct tcphdr);
          while (tcpopt < tcpdata && *tcpopt) {
            if (*tcpopt == 1)
              tcpopt++;
            else {
              int len = *(tcpopt+1);
#ifdef SF_TCP_DEBUG_TCPOPT
              printk("sf: tcp: option kind=%i, len=%i\n", *tcpopt, len);
#endif
              if (*tcpopt == 3 && len == 3)
                opt_scale = *(tcpopt+2);

              tcpopt += len;
            }
          } /* end parse options */
        } /* check_sequence */

        /* this should allow simultaneous opens */
        if (!reverse) {
          if (sftcp->state < SF_TCP_SYN_ACK) {
#ifdef SF_DEBUG_TCP
	    printk("sf: tcp: client %ssent SYN\n",
                   sftcp->state == SF_TCP_CLIENT_SYN ? "re" : "");
#endif

            if (sftcp->check_sequence) {
	      /* set initial sequence/window numbers for client */
              sftcp->client_seq = ntohl(tcp->seq);
              sftcp->client_win = window;
              if (opt_scale != -1)
                sftcp->client_scale = opt_scale;
            }

            set_tcp_timer(sftcp, SF_TCP_SYN_TO);
            check_for_fragment(ip);	/* check if packet is fragmented */

            if (sftcp->state == SF_TCP_CLIENT_SYN)
              return FW_ACCEPT;

	    sftcp->state = SF_TCP_CLIENT_SYN;
            
            if (!sftcp->log_open) 
              return FW_ACCEPT;
            else {
              returncode  =
              rule->fw_rc = FW_ACCEPT;
              goto informlog;
            }
          } /* end state < SYN_ACK */
          else { 
            /* "Again!?" [Andie McDowell as Dougs wife in the movie Multiplicity]
               --- We KNOW the server got the SYN, because we have seen the
               corresponding SYN ACK. No reason to accept this SYN! */

            return FW_BLOCK;
          }
        } /* end !reverse */
        else { /* is reverse */
          if (sftcp->state < SF_TCP_SYN_ACK) {
#ifdef SF_DEBUG_TCP
	    printk("sf: tcp: connection established\n");
#endif

            if (!sftcp->check_sequence || (ack_seq-1 == sftcp->client_seq) 
                || !tcp->ack) {
              /* first valid SYN ACK ever */

	      sftcp->state = SF_TCP_SYN_ACK;

              if (sftcp->check_sequence) {
                if (tcp->ack) /* simultaneous open kludge */
                  sftcp->client_seq++;
	
                /* set initial sequence number for server */
                sftcp->server_seq = ntohl(tcp->seq);
                sftcp->server_win = window;
                if (opt_scale != -1)
                  sftcp->server_scale = opt_scale;
                else
                  sftcp->client_scale = 0;
              }

              /* defend against SYN flooding */
              if (!sf_inside(ip->saddr))
                set_tcp_timer(sftcp, SF_TCP_SYN_TO);
              else {
                /* fabricate ACK for this SYN ACK and send it inside */
	        send_tcp(0, rif, ip, tcp);
                /* set timer to kill connection if not established */
                set_tcp_rst_timer(sftcp, SF_TCP_SYN_TO);
              }

              check_for_fragment(ip);	/* check if packet is fragmented */
              return FW_ACCEPT;

            } /* end valid SYN ACK */
            else {
#ifdef SF_DEBUG_SEQ
              printk("Sequence number of SYN ACK does not match!\n");
              printk("client_seq = %u ack_seq = %u\n", sftcp->client_seq, ack_seq);
#endif

              return FW_BLOCK;
            } /* end invalid SYN ACK */
          } /* end state < SYN_ACK */
          else {
            /* We are in SYN_ACK state (>= ESTABLISHED was caught above) */

            if (sftcp->check_sequence && (ack_seq != sftcp->client_seq)) {
#ifdef SF_DEBUG_SEQ
              printk("Sequence number of resent SYN ACK does not match!\n");
              printk("client_seq = %u ack_seq = %u\n", sftcp->client_seq, ack_seq);
#endif

              return FW_BLOCK;
            }
          } /* end SYN_ACK state */
        } /* end reverse */

        set_tcp_timer(sftcp, SF_TCP_SYN_TO);

        check_for_fragment(ip);	/* check if packet is fragmented */
        return FW_ACCEPT;
      }
      /* --- end syn --- */

    }
    /* --- end simple packets, !rst && !fin --- */

    /* --- fin packets --- */
    if (tcp->fin) {
      if (!sequence_ok) {
#ifdef SF_DEBUG_SEQ
        if (( reverse && (sftcp->state & SF_TCP_SERVER_FIN) == 0) ||
            (!reverse && (sftcp->state & SF_TCP_CLIENT_FIN) == 0))
          printk("sf: tcp: wrong sequence number [fin]!\n");
#endif
        /* else we don't klog it: it doesn't matter, we are already in 
           the FIN state (but block it anyway) */

        return FW_BLOCK;
      }

      /* Check for Kamikaze packets (also known as nastygrams, christmas
         tree packets, lamp test segments). Block them, because we are 
         neither Santa Claus nor J. Postel... [see RFC 1025 for explanation] */
      if (tcp->syn || tcp->rst) {
#ifdef SF_DEBUG_TCP
        printk("sf: tcp: Kamikaze packet detected\n");
#endif
        return FW_BLOCK;
      }

      sftcp->state |= reverse ? SF_TCP_SERVER_FIN : SF_TCP_CLIENT_FIN;

      if (sftcp->state < SF_TCP_TERMINATED) {
        /* Single FIN received. Since RFCs 793/1122
           only state we MAY support half close and SHOULD send a reset
           if the connection is closed, we DO support half close but set
           a timeout. Otherwise we might have a half dead/closed connection 
           forever. Make sure the timeout is long enough, however. 
           To support half close, we keep resetting the timeout (see above) */

        set_tcp_timer(sftcp, SF_TCP_HALFCLOSE);
#ifdef SF_DEBUG_TCP
	printk("sf: tcp: connection terminating\n");
#endif
      }
      else { 
        /* both FINs received, let final ACKs pass */

        set_tcp_timer(sftcp, SF_TCP_FIN_WAIT);
#ifdef SF_DEBUG_TCP
	printk("sf: tcp: connection properly terminated\n");
#endif
      }

      check_for_fragment(ip);	/* check if packet is fragmented */
      return FW_ACCEPT;
    }
    /* --- end fin packets --- */

    /* --- rst packets --- */
    if (tcp->rst) {
      if (!sequence_ok) {
        /* We are not going to close a connection if you cannot even
           guess the sequence number... */
#ifdef SF_DEBUG_SEQ
        printk("sf: tcp: wrong sequence number [rst]!\n");
#endif
        return FW_BLOCK;
      }

      /* Kamikaze check (see above for fin|rst) */
      if (tcp->syn) {
#ifdef SF_DEBUG_TCP
        printk("sf: tcp: Kamikaze packet detected\n");
#endif
        return FW_BLOCK;
      }

      sf_tcp_delete(sftcp);	/* RST aborts an connection immediately */

      check_for_fragment(ip);	/* check if packet is fragmented */
      return FW_ACCEPT;
    } 
    /* --- end rst packets --- */


    /* Only packets with !SYN, !FIN, !RST that belong to connections
       with state < SF_TCP_ESTABLISHED3 can arrive here. */

    /* This should never happen - but it does happen, e.g. when
       one side thinks the connection is still established and
       the other side tries to establish a new connection to the
       same port numbers. This can also happen if the connection
       has died on one end only (e.g. rebooted PC). There's no
       need to accept these strange packets, since the other
       side of the connection will not accept them. */

#ifdef SF_DEBUG_TCP
    printk("sf: tcp: connection in strange state:\n");
    if (reverse)  printk("s->c"); else printk("c->s");
    if (tcp->syn) printk(" SYN");
    if (tcp->fin) printk(" FIN");
    if (tcp->ack) printk(" ACK");
    if (tcp->rst) printk(" RST");
    printk("\n");
#endif

    return FW_BLOCK;
  }
  /* --- end check tcp --- */



checkrules:
  /* --- check rules --- */
  {
    for (rule = prio_rules; rule != NULL; rule = rule->fw_next) {
#ifdef SF_DEBUG_RULES
      printk("sf: rule %li, flags %lx, prot %i\n", 
             rule->rule_id.num, rule->fw_flags, rule->protocol);
#endif

      /* check if protocols match */

      if ((rule->fw_flags & SF_FW_PROT) != SF_FW_PROT_ALL) {
	if ((rule->fw_flags & SF_FW_PROT) == SF_FW_CHECK_PROTOCOL) {
	  unsigned long flg = rule->fw_flags & SF_TYPE_MASK;

	  if (rule->protocol != ip->protocol)
	    continue;
	  switch (ip->protocol) {
	    case IPPROTO_ICMP:
	      if (flg == SF_ICMP_ALLTYPES)
		break;
	      switch (icmp->type) {	/* filter msg type */
		case ICMP_ECHOREPLY:
		  if (!(flg & SF_ICMP_ECHOREPLY))
		    continue;
		  break;
		case ICMP_DEST_UNREACH:
		  if (!(flg & SF_ICMP_DEST_UNREACH))
		    continue;
		  break;
		case ICMP_SOURCE_QUENCH:
		  if (!(flg & SF_ICMP_SOURCE_QUENCH))
		    continue;
		  break;
		case ICMP_REDIRECT:
		  if (!(flg & SF_ICMP_REDIRECT))
		    continue;
		  break;
		case ICMP_ECHO:
		  if (!(flg & SF_ICMP_ECHO))
		    continue;
		  break;
		case ICMP_TIME_EXCEEDED:
		  if (!(flg & SF_ICMP_TIME_EXCEEDED))
		    continue;
		  break;
		case ICMP_PARAMETERPROB:
		  if (!(flg & SF_ICMP_PARAMETERPROB))
		    continue;
		  break;
		case ICMP_TIMESTAMP:
		  if (!(flg & SF_ICMP_TIMESTAMP))
		    continue;
		  break;
		case ICMP_TIMESTAMPREPLY:
		  if (!(flg & SF_ICMP_TIMESTAMPREPLY))
		    continue;
		  break;
		case ICMP_INFO_REQUEST:
		  if (!(flg & SF_ICMP_INFO_REQUEST))
		    continue;
		  break;
		case ICMP_INFO_REPLY:
		  if (!(flg & SF_ICMP_INFO_REPLY))
		    continue;
		  break;
		case ICMP_ADDRESS:
		  if (!(flg & SF_ICMP_ADDRESS))
		    continue;
		  break;
		case ICMP_ADDRESSREPLY:
		  if (!(flg & SF_ICMP_ADDRESSREPLY))
		    continue;
		  break;
		default:
		  continue;
	      }
	      break;
	    case IPPROTO_IGMP:
	      if (flg == SF_IGMP_ALLTYPES)
		break;
	      switch (igmp->type) {	/* filter msg type */
		case IGMP_HOST_MEMBERSHIP_QUERY:
		  if (!(flg & IGMP_HOST_MEMBERSHIP_QUERY))
		    continue;
		  break;
		case IGMP_HOST_MEMBERSHIP_REPORT:
		  if (!(flg & SF_IGMP_HOST_MEMBERSHIP_REPORT))
		    continue;
		  break;
		case IGMP_HOST_LEAVE_MESSAGE:
		  if (!(flg & IGMP_HOST_LEAVE_MESSAGE))
		    continue;
		  break;
		default:
		  continue;
	      }
	      break;
	    default:
	      break;
	  }			/* ip->protocol */
	}
	else {			/* rule->fw_flags & SF_FW_PROT != SF_FW_CHECK_PROTOCOL */
	  switch (rule->fw_flags & SF_FW_PROT) {
	    case SF_FW_PROT_RIP:
	      if ((ip->protocol != IPPROTO_UDP) || (udp->dest != htons(SF_RIP_PORT)))
		continue;
	      if (rule->fw_rip_cnt == 0)
		break;
	      if (!check_rip(rule, (struct riphdr *) (udp + 1),
			     ip_tot_len - ip_hdr_len - 8))
		continue;
	      break;
	    default:
	      continue;
	  }
	}			/* if */
      }				/* if */

      if (rule->fw_flags & SF_FW_CHECK_TTL)
	switch (rule->fw_flags & SF_FW_TTL) {
	  case SF_FW_TTL_EQUAL:
	    if (ip->ttl != rule->ttl)
	      continue;
	    break;
	  case SF_FW_TTL_LESS:
	    if (ip->ttl >= rule->ttl)
	      continue;
	    break;
	  case SF_FW_TTL_GREATER:
	    if (ip->ttl <= rule->ttl)
	      continue;
	    break;
	  case SF_FW_TTL_NOTEQUAL:
	    if (ip->ttl == rule->ttl)
	      continue;
	    break;
	  default:
	    continue;
	}

      if (rule->fw_flags & SF_FW_CHECK_OPT) {
	unsigned char *buff = (unsigned char *) (ip + 1);	/* start of options */
	int done = 0;
	int len = sizeof(struct iphdr);

	while (!done && len < ip_hdr_len)	/* end of header not yet reached */
	  switch (*buff) {
	    case IPOPT_END:	/* pad bytes */
#ifdef SF_DEBUG_OPT
	      printk("option end\n");
#endif
	      done = 1;
	      break;		/* end of options */
	    case IPOPT_NOOP:
#ifdef SF_DEBUG_OPT
	      printk("option noop\n");
#endif
	      buff++;
	      len++;
	      break;
	    case IPOPT_SEC:	/* security */
#ifdef SF_DEBUG_OPT
	      printk("option sec\n");
#endif
	      if (rule->fw_flags & SF_FW_OPT_SEC) {
		done = 2;
		break;
	      }
	      buff += 11;
	      len += 11;
	      break;
	    case IPOPT_LSRR:
#ifdef SF_DEBUG_OPT
	      printk("option lsr\n");
#endif
	      if (rule->fw_flags & SF_FW_OPT_LSR) {
		done = 2;
		break;
	      }
	      buff++;
	      len  += *buff;
	      buff += (*buff - 1);
	      break;
	    case IPOPT_SSRR:
#ifdef SF_DEBUG_OPT
	      printk("option ssr\n");
#endif
	      if (rule->fw_flags & SF_FW_OPT_SSR) {
		done = 2;
		break;
	      }
	      buff++;
	      len  += *buff;
	      buff += (*buff - 1);
	      break;
	    case IPOPT_RR:
#ifdef SF_DEBUG_OPT
	      printk("option rr\n");
#endif
	      if (rule->fw_flags & SF_FW_OPT_RR) {
		done = 2;
		break;
	      }
	      buff++;
	      len  += *buff;
	      buff += (*buff - 1);
	      break;
	    case IPOPT_SID:
#ifdef SF_DEBUG_OPT
	      printk("option sid\n");
#endif
	      if (rule->fw_flags & SF_FW_OPT_SATID) {
		done = 2;
		break;
	      }
	      buff += 4;
	      len  += 4;
	      break;
	    case IPOPT_TIMESTAMP:
#ifdef SF_DEBUG_OPT
	      printk("option ts\n");
#endif
	      if (rule->fw_flags & SF_FW_OPT_TS) {
		done = 2;
		break;
	      }
	      buff++;
	      len  += *buff;
	      buff += (*buff - 1);
	      break;
	    default:		/* unknown option */
#ifdef SF_DEBUG_OPT
	      printk("unknown option %i\n", *buff);
#endif
	      done = 1;
	      break;
	  }
#ifdef SF_DEBUG_OPT
	printk(" done options = %i\n", done);
#endif
	if (done != 2)
	  continue;		/* no option matched */
      }

      /* check addresses and ports */
      if (rule->fw_src_cnt) {
	int match = 0;
	int port  = -1;

	if ((ip->protocol == IPPROTO_TCP) || (ip->protocol == IPPROTO_UDP))
	  port = ntohs(tcp->source);
	match = sf_addr_match(rule->fw_src_idx, rule->fw_src_cnt, 
                              ip->saddr, port, rule->fw_flags & SF_FW_SRC_NEG);
	if ((match && (rule->fw_flags & SF_FW_SRC_NEG))
	    || (!match && !(rule->fw_flags & SF_FW_SRC_NEG)))
	  continue;
      }
      if (rule->fw_dst_cnt) {
	int match = 0;
	int port  = -1;

	if ((ip->protocol == IPPROTO_TCP) || (ip->protocol == IPPROTO_UDP))
	  port = ntohs(tcp->dest);
	match = sf_addr_match(rule->fw_dst_idx, rule->fw_dst_cnt, 
                              ip->daddr, port, rule->fw_flags & SF_FW_DST_NEG);
	if ((match && (rule->fw_flags & SF_FW_DST_NEG))
	    || (!match && !(rule->fw_flags & SF_FW_DST_NEG)))
	  continue;
      }

#ifdef SF_DEBUG_RULES
      printk("sf: packet matches rule\n");
#endif
      if (rule->fw_rc == SF_RC_OBSERVE)
        inform_log(rule, ip, rif, ip_tot_len);
      else
        break;			/* the packet matches the rule */
    }
  }				/* check rules */

  if (rule == NULL)
    return FW_BLOCK;		/* packet matches no rule */

  returncode = rule->fw_rc;

  if (rule->fw_rc == FW_ACCEPT) {	

    if (ip->protocol == IPPROTO_TCP) {
      /* if a tcp packet arrives here, it is initiating a new, legal connection */
      sftcp = sf_tcp_insert(ip->saddr, ip->daddr, tcp->source, tcp->dest, rule, rif);

      if (sftcp != NULL) {
        if (rule->fw_flags & SF_FW_LOG)
          sftcp->log_open = 1;
        if (rule->fw_flags & SF_FTP_DATA_LOG)
          sftcp->log_dxfer = 1;
        if (!(rule->fw_flags & SF_FTP_NO_ACTIVE))
          sftcp->ftp_allow_active = 1;
        if (!(rule->fw_flags & SF_FTP_NO_PASSIVE))
          sftcp->ftp_allow_passive = 1;
	if (rule->fw_flags & SF_CHECK_SEQUENCE)
	  sftcp->check_sequence = 1;
        goto have_connection;
      }

      return FW_BLOCK; /* internal error */
    } /* TCP */

    check_for_fragment(ip);	/* check if packet is fragmented */
  }
  else if (rule->fw_rc == SF_RC_TREJECT) {	/* TCP reset / best reject */
    if (opt == SF_STATE_TRANSMIT)
      returncode = FW_REJECT;
    else {			/* choose best method */
      /* Note: This branch never gets called for local packets, because 
         local packets are checked at STATE_TRANSMIT only. We would not
         be able to receive RST packets anyway, because the filter would 
         block them, as they are not on the connection list... */

      if (ip->protocol == IPPROTO_UDP) {	/* send ICMP port unreach. */
	send_icmp_error(rif, ip, ICMP_DEST_UNREACH, ICMP_PORT_UNREACH);
	returncode = FW_BLOCK;
      }
      else if (ip->protocol == IPPROTO_TCP) {	/* send TCP reset */
	send_tcp(1, rif, ip, tcp);
	returncode = FW_BLOCK;	/* block packet */
      }
      else			/* simple reject */
	returncode = FW_BLOCK;
    }
  }
  else if (rule->fw_rc == SF_RC_ECHO) {
    if ((ip->protocol == IPPROTO_ICMP) && (icmp->type == ICMP_ECHO))
      icmp_reflect(rif, ip, icmp);
    returncode = FW_BLOCK;	/* in any case, block the packet */
  }
  else if (rule->fw_rc < FW_REJECT)	/* Filter ICMP options */
    if (opt == SF_STATE_TRANSMIT)
      returncode = FW_REJECT;
    else {
      /* Linux FW functions are limited -- do it yourself... */
      switch (rule->fw_rc) {
	case SF_RC_RNET:
	  send_icmp_error(rif, ip, ICMP_DEST_UNREACH, ICMP_NET_UNREACH);
	  break;
	case SF_RC_RPROTO:
	  send_icmp_error(rif, ip, ICMP_DEST_UNREACH, ICMP_PROT_UNREACH);
	  break;
	case SF_RC_RPORT:
	  send_icmp_error(rif, ip, ICMP_DEST_UNREACH, ICMP_PORT_UNREACH);
	  break;
      }
      returncode = FW_BLOCK;
    }

  rule->bytes += ip_tot_len;
  if (rule->bytes >= GIGA) {
    rule->gbytes++;
    rule->bytes -= GIGA;
  }

informlog:
  if (inform_log(rule, ip, rif, ip_tot_len) < 0)
    return FW_BLOCK;

  /* return rule dependent return code */
  return returncode;
}


/* 
 * initialization and interface functions
 */

int sf_init(void)
{
  int i;
  unsigned long flags;

#ifdef SF_DEBUG
  printk("sf_init() called\n");
#endif

  save_flags(flags);
  cli();

  /* initialize rules/conns lists */
  prio_rules =
  rules      =
  rule_first_next = NULL;

  conns =
  lastc = NULL;
  conn_id = 0;

#ifndef CONFIG_IP_ALWAYS_DEFRAG
  /* initialize fragment hash queues */
  for (i = 0; i < SF_FRAG_HASH_SIZE; i++)
    sf_frag[i] = NULL;
#endif

  /* initialize tcp hash queues */
  for (i = 0; i < SF_TCP_HASH_SIZE; i++)
    sf_tcp[i] = NULL;

  sf_fw_initialized = 1;
  restore_flags(flags);

  return 0;
}

int sf_resetusage(void)
{
  unsigned long flags;
  struct sf_fw *rule;

#ifdef SF_DEBUG
  printk("sf_resetusage() called\n");
#endif

  save_flags(flags);
  cli();

  rule = prio_rules;

  while (rule != NULL) {
    rule->usage  =
    rule->bytes  =
    rule->gbytes = 0;
    rule = rule->fw_next;
  }

  restore_flags(flags);
  return 0;
}


int sf_clear(void)
{
  unsigned long flags;
  struct sf_fw *rule;

#ifdef SF_DEBUG
  printk("sf_clear() called\n");
#endif

  save_flags(flags);
  cli();

  while (prio_rules != NULL) {
    rule = prio_rules;
    prio_rules = prio_rules->fw_next;
    kfree_s(rule, sizeof(struct sf_fw));
  }
  rules = NULL;

  restore_flags(flags);
  return 0;
}

int sf_add(struct sf_fw *rule, int priority)
{
  unsigned long flags;
  struct sf_fw *r;

#ifdef SF_DEBUG
  printk("sf_add() called, adding rule %li\n", rule->rule_id.num);
#endif

  save_flags(flags);
  cli();

  if (priority) {
    rule->fw_next = prio_rules;
    prio_rules = rule;
  }
  else {
    if (prio_rules != rules) {
      for (r = prio_rules; (r != NULL) && (r->fw_next != rules); r = r->fw_next);
      if (r != NULL)
        r->fw_next = rule;
      else
        prio_rules = rule;
    }
    else
      prio_rules = rule;
    rule->fw_next = rules;
    rules = rule;
  }

  restore_flags(flags);
  return 0;
}

int sf_replace(struct sf_fw *rule)
{
  unsigned long flags;
  struct sf_fw *rp1, *rp2;

#ifdef SF_DEBUG
  printk("sf_replace() called, replacing rule %li\n", rule->rule_id.num);
#endif

  save_flags(flags);
  cli();

  for (rp1 = prio_rules; 
       (rp1 != NULL) && (rp1->rule_id.num != rule->rule_id.num);
       rp1 = rp1->fw_next);

  if (rp1 == NULL) {		/* id not found */
    restore_flags(flags);
    return -1;
  }
  rp2 = rp1->fw_next;
  memcpy(rp1, rule, sizeof(struct sf_fw));
  rp1->fw_next = rp2;

  kfree_s(rule, sizeof(struct sf_fw));

  restore_flags(flags);
  return 0;
}

int sf_delete(unsigned long id)
{
  unsigned long flags;
  struct sf_fw *rp1, *rp2;

#ifdef SF_DEBUG
  printk("sf_delete() called, deleting rule %lu\n", id);
#endif

  save_flags(flags);
  cli();

  rp1 = prio_rules;
  rp2 = NULL;
  while ((rp1 != NULL) && (rp1->rule_id.num != id)) {
    rp2 = rp1;
    rp1 = rp1->fw_next;
  }

  if (rp1 == NULL) {		/* id not found */
    restore_flags(flags);
    return -1;
  }

  if (rp2 != NULL)
    rp2->fw_next = rp1->fw_next;
  else
    prio_rules = rp1->fw_next;
  if (rp1 == rules)
    rules = rp1->fw_next;

  kfree_s(rp1, sizeof(struct sf_fw));

  restore_flags(flags);
  return 0;
}

int sf_flush(void)
{
  struct sf_fw *rule;
  struct sf_tcp_connection *sftcp;
  unsigned long flags;

  save_flags(flags);
  cli();

#ifdef SF_DEBUG
  printk("sf_flush() called\n");
#endif

  sftcp = conns;
  while (sftcp != NULL) {
    unsigned short source = ntohs(sftcp->client_port);
    unsigned short dest   = ntohs(sftcp->server_port);

    for (rule = prio_rules; rule != NULL; rule = rule->fw_next) {

      if ((rule->protocol != IPPROTO_TCP)
	  && ((rule->fw_flags & SF_FW_PROT) != SF_FW_PROT_ALL))
	continue;
      if (rule->fw_src_cnt) {
	int match = sf_addr_match(rule->fw_src_idx, rule->fw_src_cnt,
	     sftcp->client_addr, source, rule->fw_flags & SF_FW_SRC_NEG);

	if ((match && (rule->fw_flags & SF_FW_SRC_NEG))
	    || (!match && !(rule->fw_flags & SF_FW_SRC_NEG)))
	  continue;
      }
      if (rule->fw_dst_cnt) {
	int match = sf_addr_match(rule->fw_dst_idx, rule->fw_dst_cnt,
	       sftcp->server_addr, dest, rule->fw_flags & SF_FW_DST_NEG);

	if ((match && (rule->fw_flags & SF_FW_DST_NEG))
	    || (!match && !(rule->fw_flags & SF_FW_DST_NEG)))
	  continue;
      }
      break;			/* rule matches */
    }

    if ((rule == NULL) || (rule->fw_rc != FW_ACCEPT)) {
      struct sf_tcp_connection *del = sftcp;

      sftcp = sftcp->after;
      sf_tcp_delete(del);
    }
    else
      sftcp = sftcp->after;
  }

  restore_flags(flags);
  return 0;
}

int sf_flush_all(void)
{
  struct sf_tcp_connection *sftcp, *del;
  unsigned long flags;

  save_flags(flags);
  cli();

#ifdef SF_DEBUG
  printk("sf_flush_all() called\n");
#endif

  sftcp = conns;
  while (sftcp != NULL) {
    del = sftcp;

#ifdef SF_DEBUG_TCP
    printk("sf: killing %p ", sftcp);
    PRINTIP(sftcp->client_addr);
    printk(":%u -> ", ntohs(sftcp->client_port));
    PRINTIP(sftcp->server_addr);
    printk(":%u\n", ntohs(sftcp->server_port));
#endif

    sftcp = sftcp->after;
    sf_tcp_delete(del);
  }

  restore_flags(flags);
  return 0;
}

int sf_tcp_kill(struct sf_control_tcp *ctl)
{
  struct sf_tcp_connection *sftcp;
  unsigned long flags;
  int rc = -1;

  save_flags(flags);
  cli();

  if (ctl->key < SF_TCP_HASH_SIZE) {	/* sanity check */
    sftcp = sf_tcp[ctl->key];
    while (sftcp != NULL && (sftcp->id != ctl->id ||
			     sftcp->established != ctl->established))
      sftcp = sftcp->next;

    if (sftcp != NULL) {
      /* Found connection, kill it */
      sf_tcp_delete(sftcp);
      rc = 0;
    }
  }

  restore_flags(flags);
  return rc;
}

void sf_del_all_timers(void)
{
#ifndef CONFIG_IP_ALWAYS_DEFRAG
  struct sf_fragment *frag;
  int i;
#endif
  struct sf_tcp_connection *sftcp;
  unsigned long flags;

  save_flags(flags);
  cli();

#ifndef CONFIG_IP_ALWAYS_DEFRAG
  for (i = 0; i < SF_FRAG_HASH_SIZE; i++)
    while ((frag = sf_frag[i]) != NULL) {
      del_timer(&frag->timer);
      sf_frag[i] = frag->next;
      kfree_s(frag, sizeof(struct sf_fragment));
    }
#endif

  sftcp = conns;
  while (sftcp != NULL) {
    if (sftcp->timer_set) {
      struct sf_tcp_connection *del = sftcp;

      sftcp = sftcp->after;
      sf_tcp_delete(del);
    }
    else
      sftcp = sftcp->after;
  }

  restore_flags(flags);
}

int sf_write_config(const char *buffer, int len)
{
  struct sf_control *control;
  struct sf_control_tcp *ctl;
  struct sf_fw *rule;
  int rc, i;

  control = (struct sf_control *) kmalloc(len, GFP_ATOMIC);

  memcpy_fromfs(control, buffer, len);

  if (control->magic != SF_CONTROL_MAGIC) {
    kfree_s(control, len);
    return -EINVAL;
  }

  if (!sf_fw_initialized)
    sf_init();

  switch (control->command) {
    case SF_COMMAND_DISABLE:
      sf_fw_enabled = 0;
      rc = len;
      break;

    case SF_COMMAND_ENABLE:
      sf_log_data.num_entries = 0;
      sf_log_data.lost_entries = 0;
      sf_fw_enabled = 1;
      rc = len;
      break;

    case SF_COMMAND_CLEAR_LOST:
      sf_log_data.lost_entries = 0;
      rc = len;
      break;

    case SF_COMMAND_FW_CLEAR:
      if (sf_clear() != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    case SF_COMMAND_FW_ADDPRIO:
    case SF_COMMAND_FW_ADD:
      if (len != (sizeof(struct sf_control) + sizeof(struct sf_fw))) {
	rc = -EINVAL;
	break;
      }
      if ((rule = (struct sf_fw *) kmalloc(sizeof(struct sf_fw), GFP_ATOMIC)) == NULL) {
	rc = -ENOMEM;
	break;
      }
      memcpy(rule, control->data, sizeof(struct sf_fw));

      if (sf_add(rule, control->command == SF_COMMAND_FW_ADDPRIO) != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    case SF_COMMAND_FW_REPLACE:
      if (len != (sizeof(struct sf_control) + sizeof(struct sf_fw))) {
	rc = -EINVAL;
	break;
      }
      if ((rule = (struct sf_fw *) kmalloc(sizeof(struct sf_fw), GFP_ATOMIC)) == NULL) {
	rc = -ENOMEM;
	break;
      }
      memcpy(rule, control->data, sizeof(struct sf_fw));

      if (sf_replace(rule) != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    case SF_COMMAND_FW_DELETE:
      if (len != (sizeof(struct sf_control) + sizeof(unsigned long))) {
	rc = -EINVAL;
	break;
      }
      if (sf_delete(*(unsigned long *) (control->data)) != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    case SF_COMMAND_FW_ADDR:
#ifdef SF_DEBUG
      printk("sf_addr: transmitted address array\n");
#endif
      if (len != (sizeof(struct sf_control) + sizeof(sf_addr))) {
	rc = -EINVAL;
	break;
      }
      memcpy(sf_addr, control->data, sizeof(sf_addr));
      sf_addr_free = sf_addr[0].mask;

      /* initialize interface cache hash table */
      for (i = 0; i < SF_IF_CACHE_SIZE; i++)
	sf_if_cache[i].ifaddr = 0;

      rc = len;
      break;

    case SF_COMMAND_FLUSH:
      if (sf_flush() != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    case SF_COMMAND_FLUSH_ALL:
      if (sf_flush_all() != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    case SF_COMMAND_RESETUSAGE:
      if (sf_resetusage() != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    case SF_COMMAND_TCP_KILL:
      if (len != (sizeof(struct sf_control) + sizeof(struct sf_control_tcp))) {
	rc = -EINVAL;
	break;
      }
      ctl = (struct sf_control_tcp *) control->data;
      if (sf_tcp_kill(ctl) != 0) {
	rc = -1;
	break;
      }
      rc = len;
      break;

    default:
      rc = -EINVAL;
      break;
  }

  kfree_s(control, len);
  return rc;

}


int sf_forward_chk(struct firewall_ops *this,
		   int pf,
		   struct device *dev,
		   void *phdr,
		   void *arg)
{
  return sf_fw_chk(phdr, dev, SF_STATE_FORWARD);
}


int sf_input_chk(struct firewall_ops *this,
		 int pf,
		 struct device *dev,
		 void *phdr,
		 void *arg)
{
  return sf_fw_chk(phdr, dev, SF_STATE_RECEIVE);
}


int sf_output_chk(struct firewall_ops *this,
		  int pf,
		  struct device *dev,
		  void *phdr,
		  void *arg)
{
  return sf_fw_chk(phdr, dev, SF_STATE_TRANSMIT);
}

struct firewall_ops sf_fw_ops =
{
  NULL,
  sf_forward_chk,
  sf_input_chk,
  sf_output_chk,
  PF_INET,
  2				/* higher priority than std. firewall */
};
